//////////////////////////////////////////////////////////////////////////////
//
//      Copyright  1999 Zentropic Computing, All rights reserved
//
// Authors:             Steve Papacharalambous
// Original date:       Sun 01 Aug 1999
//
// pthreads interface for Real Time Linux.
//
//
// The contents of this file are subject to the Mozilla Public
// License Version 1.0 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

//
// pthreads mutex test program.
//
// This program tests the mutex operations.
// At module initialisation a mutex is created and locked.  A low priority control
// thread is created (SCHED_OTHER/0) which tries to lock the mutex.  NUM_THREADS threads are then
// created with increrasing priorities (SCHED_FIFO) which also try to gain the mutex.
//  The mutex queue is displayed to verify that the tasks are queued correctly and
// that tasks are inserted into the mutex queue in order of increasing priority.
// The initialisation function then releases the mutex and the threads insert data
// into a data buffer using the mutex to gain atomic access.  The contents of the
// data buffer are displayed when the test module is removed to verify that the data
// has been  inserted into the buffer is the expected order.
//






#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>

#include <rtai_pthread.h>



#define RUNNABLE_ON_CPUS 3
#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define TIMER_TO_CPU 3             // < 0 || > 1 to maintain a symmetrically processed timer

#define RT_STACK 2000

#define DEFAULT_TICK_PERIOD 20000

#define T_BUFFER_SIZE 256

#define NUM_THREADS 4

struct op_buffer {
  int t_buffer_data[T_BUFFER_SIZE];
  long t_buffer_tid[T_BUFFER_SIZE];
  int t_buffer_pos;
  pthread_mutex_t lock;
};

static struct op_buffer test_op;


void *thread_func(void *arg) {

  int i;
  int r_c;
  int my_policy;
  pthread_t self;
  struct sched_param my_param;


  self = pthread_self();

// Set thread scheduling parameters.
  my_param.sched_priority = 20 + (int)self;
  if(( r_c = pthread_setschedparam(self, SCHED_FIFO, &my_param)) != 0) {
    printk("pthread: %s, pthread_setschedparam error: %d\n", (char *)arg, r_c);
  }

  if(( r_c = pthread_getschedparam(self, &my_policy, &my_param)) != 0) {
    printk("pthread: %s, pthread_getschedparam error: %d\n", (char *)arg, r_c);
  }

  printk("pthread: %s, running at %s/%d\n", (char *)arg,
          (my_policy == SCHED_FIFO ? "FIFO"
          : (my_policy == SCHED_RR ? "RR"
          : (my_policy == SCHED_OTHER ? "OTHER"
          : "unknown"))),
          my_param.sched_priority);



  printk("Starting pthread: %s, id: %ld\n", (char *)arg, self);

  if(( r_c = pthread_mutex_lock(&test_op.lock)) != 0) {
    printk("pthread %s, Cannot gain mutex: %d\n", (char *)arg, r_c);
  }


  for(i = 0; i < 5; i++) {

    printk("pthread: %s, loop count %d\n", (char *)arg, i);
    test_op.t_buffer_data[test_op.t_buffer_pos] = self + i;
    test_op.t_buffer_tid[test_op.t_buffer_pos] = self;
    test_op.t_buffer_pos++;
  }

  if(( r_c = pthread_mutex_unlock(&test_op.lock)) != 0) {
    printk("pthread %s - Error unlocking mutex: %d\n", (char *)arg, r_c);
  }

  pthread_exit("b");

  return(NULL);

}


void *control_func(void *arg) {

  int i;
  int r_c;
  int my_policy;
  pthread_t self;
  struct sched_param my_param;


  self = pthread_self();

  printk("Starting pthread: %s, id: %ld\n", (char *)arg, self);

  if(( r_c = pthread_getschedparam(self, &my_policy, &my_param)) != 0) {
    printk("pthread: %s, pthread_getschedparam error: %d\n", (char *)arg, r_c);
  }

  printk("pthread: %s, running at %s/%d\n", (char *)arg,
          (my_policy == SCHED_FIFO ? "FIFO"
          : (my_policy == SCHED_RR ? "RR"
          : (my_policy == SCHED_OTHER ? "OTHER"
          : "unknown"))),
          my_param.sched_priority);


  if(( r_c = pthread_mutex_lock(&test_op.lock)) != 0) {
    printk("pthread %s, Cannot gain mutex: %d\n", (char *)arg, r_c);
  }


  if(( r_c = pthread_mutex_unlock(&test_op.lock)) != 0) {
    printk("pthread %s - Error unlocking mutex: %d\n", (char *)arg, r_c);
  }

  pthread_exit("b");

  return(NULL);

}


int init_module(void) {

  int r_c, i;
  pthread_t thread_id = 0;
  RTIME tick_period;
  QUEUE *mutex_q;
  RT_TASK *rt_task;

  printk("\nmutex test program.\n");

// Initialise & lock the mutex: test_op
  pthread_mutex_init(&test_op.lock, NULL);
  test_op.t_buffer_pos = 0;
  for(i = 0; i < T_BUFFER_SIZE; i++) {
    test_op.t_buffer_data[i] = 0;
    test_op.t_buffer_tid[i] = 0;
  }

  if(( r_c = pthread_mutex_lock(&test_op.lock)) != 0) {
    printk("mutextest - Error creating mutex: %d\n", r_c);
    return(-1);
  }

// Create a control pthread.
  if(( r_c = pthread_create(&thread_id, NULL, control_func, "control pthread mutex tester")) != 0 ) {

    printk("mutextest - Control thread creation failed:  %d\n", r_c);
    return(-1);
  }

// Create NUM_THREADS application pthreads.
  for(i = 0; i < NUM_THREADS; i++) {

    if(( r_c = pthread_create(&thread_id, NULL, thread_func, "application pthread mutex tester")) != 0 ) {

      printk("mutextest - Application thread creation failed:  %d\n", r_c);
      return(-1);
    }

    printk("mutextest - Created pthread, id = %ld\n", thread_id);

  }

// Print the contents of the mutex queue.
  printk("\nmutextest - Contents of the mutex queue.....\n");
  printk("-------------------<Start>--------------------\n");
  mutex_q = &(test_op.lock.m_semaphore.queue);
  while((mutex_q = mutex_q->next) != &(test_op.lock.m_semaphore.queue)) {

    printk("Priority: %d, Base priority: %d\n",(mutex_q->task)->base_priority , (mutex_q->task)->priority);

  }

  printk("--------------------<End>---------------------\n\n");


// Unlock the threads & let em rip!
  if(( r_c = pthread_mutex_unlock(&test_op.lock)) != 0) {
    printk("mutextest - Error unlocking mutex: %d\n", r_c);
  }

  return(0);

}


void cleanup_module(void) {

  int r_c;
  int i;

  printk("mutextest - Contents of mutex test_op buffer:\n");
  for(i = 0; i < test_op.t_buffer_pos; i++){
    printk("Position: %d, Value: %d, Thread id: %ld\n", i, test_op.t_buffer_data[i], test_op.t_buffer_tid[i]);
  }

  if(( r_c = pthread_mutex_destroy(&test_op.lock)) != 0) {
    printk("mutextest - Error removing mutex: %d\n", r_c);
  }

  printk("mutex test module removed.\n");

}
