//////////////////////////////////////////////////////////////////////////////
//
//      Copyright  1999 Zentropic Computing, All rights reserved
//
// Authors:             Steve Papacharalambous
// Original date:       Fri 27 Aug 1999
//
// Utility routines for Real Time Linux.
//
//
// The contents of this file are subject to the Mozilla Public
// License Version 1.0 (the "License"); you may not use this file
// except in compliance with the License. You may obtain a copy of
// the License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS
// IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
// implied. See the License for the specific language governing
// rights and limitations under the License.
//
///////////////////////////////////////////////////////////////////////////////

static char id_rtai_utils_c[] __attribute__ ((unused)) = "@(#)$Id: rtai_utils.c,v 1.1.1.1 1999/08/31 14:12:20 stevepapa Exp $";


#include <linux/vmalloc.h>
#include <linux/errno.h>

// ----------------------------------------------------------------------------

#ifndef NULL
#define NULL ((void *) 0)
#endif

static unsigned int alloc_size = 0;
static unsigned int chunk_count = 0;
static char *rt_mem_ptr = NULL;
static char *rt_mem_stat_ptr = NULL;

// ----------------------------------------------------------------------------

// RT memory allocation functions.
//
// These routines provide a very simple form of memory allocation which
// can by used by RT tasks that cannot be blocked or subjected to
// significant delays during memory allocation.
//
// Currently only fixed memory chunks of preset size are allocated, requests
// for memory sizes greater than the preset size are rejected.   Requests
// for memory sizes less than or equal to the preset size are each allocated
// a chunk of this memory providing there is a free one available in the pool.
//
// These limits, the memory chunk size and maximum number of memory chunks
// in the pool are set by the application by calling rt_mem_init
// The application must call rt_mem_init once to initialise the memory
// pool.  It can then use the functions rt_malloc and rt_free to dynamically
// allocate and free memory during RT operation.
// It must call rt_mem_end on termination so that the memory can
// be cleaned up.
//


// Memory allocation system initialisation.
int rt_mem_init(unsigned int chunk_size, unsigned int num_chunks) {

  int i;

  if(num_chunks <= 0 || chunk_size <= 0) {
    return(-EINVAL);
  }

  rt_mem_ptr = (char *)vmalloc(chunk_size * num_chunks);
  if(!rt_mem_ptr) {
    return(-ENOMEM);
  }

  memset((void *)rt_mem_ptr, 0, (chunk_size * num_chunks));

  rt_mem_stat_ptr = (char *)vmalloc(num_chunks);
  if(!rt_mem_ptr) {
    return(-ENOMEM);
  }

  memset((void *)rt_mem_stat_ptr, 0, num_chunks);

  alloc_size = chunk_size;
  chunk_count = num_chunks;

  for(i = 0; i < chunk_size * num_chunks; i++) {
    *(rt_mem_ptr + i) = 0;
  }

  for(i = 0; i < num_chunks; i++) {
    *(rt_mem_stat_ptr + i) = 0;
  }

  return(0);

}  // End function - rt_mem_init


// Memory allocation.
void *rt_malloc(int size) {

  int i;
  int found_free_chunk;

  if(size > alloc_size) {
    return(NULL);
  }

  found_free_chunk = 0;
  for(i = 0; i < chunk_count; i++) {
    if(*(rt_mem_stat_ptr + i) == 0) {
      found_free_chunk = 1;
      break;
    }
  }

  if(found_free_chunk) {
    *(rt_mem_stat_ptr + i) = 1;
    return((void *)(rt_mem_ptr + (i * alloc_size)));
  }

  return(NULL);

}  // End function - rt_malloc


// Memory release.
void rt_free(void *mem_ptr) {

  unsigned int mem_index;

  if(mem_ptr == NULL) {
    return;
  }

  mem_index = ((unsigned int)mem_ptr - (unsigned int)rt_mem_ptr) / alloc_size;
  if(mem_index > chunk_count) {
    return;
  }

  *(rt_mem_stat_ptr + mem_index) = 0;
  return;

}  // End function - rt_free

// Memory allocation system termination.
void rt_mem_end(void) {
  vfree((void *)rt_mem_ptr);
  vfree((void *)rt_mem_stat_ptr);
}  // End function - rt_mem_end


// ---------------------------------< eof >------------------------------------
